(***********************************************************************)
(*                                                                     *)
(*                              Ref-image                              *)
(*                                                                     *)
(*                  Image Segmentation by Refinement                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2009-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: segmentation.ml,v 1.62 2014-02-26 00:12:42 fclement Exp $ *)

module Nat = Unatural;;
module Dgraph = Udynamic_graph;;
module Ddgraph = Udelayed_dgraph;;

module Set = Umatrix_nonempty_set;;
module Mapping = Umatrix_graph;;
module Structural_mapping = Umatrix_structural_mapping;;
module Config = Configuration;;

let s_qual = Upervasives.qualify "Segmentation";;

let s_failwith = Upervasives.failwith_qual s_qual;;
let s_print_debug fname = Config.print_debug (s_qual fname);;

open Segmentation_types;;
open Segmentation_spec;;
open Image;;
open Zonation;;
open Configuration_types;;

module Make =
  functor (Scal : SCALAR) ->
  functor (Cod : CODOMAIN) ->
struct
  type scalar = Scal.t;;
  type 'a codomain = 'a Cod.t;;
  
  module St = State.Make (Scal) (Cod);;
  module H = History.Make (Config) (St);;
  
  let print_iteration_report ppf i j =
    Monitor.print_iteration_report Scal.print ppf (i, j);
  ;;
  
  let print_zone_splitting_report ppf
      best_zn pim_ip1 szonef paramf szonet paramt =
    let znf = best_zn and znt = last_zone_number pim_ip1
    and cardf = Set.cardinal szonef.Set.nonempty_set
    and cardt = Set.cardinal szonet.Set.nonempty_set in
    Config.print_report
      (Monitor.print_zone_splitting_report Scal.print) ppf
      (best_zn, (znf, cardf, paramf), (znt, cardt, paramt))
  ;;
  
  let _print_gimage = Segmentation_types_print.print_gimage Scal.print;;
  
  (* Early parse of the command line to get the image size before setting up
   graphical capabilities when applying Display.Make functor. *)
  let zsh_0 = H.initialization ();;
  
  module Disp = Display.Make (Scal) (Cod) (St) (H);;
  
  let ppf = Config.get_formatter ();;
  
  (* Iteration. *)
  
  (* Steps 1a and 1b of the multiscalar algorithm, pp. 13-14. RR INRIA 7446. *)
  (* scalar zonation codomain -> scalar zonation -> scalar zonation *)
  let scalar_cutting_candidates
      zs
      ({ other_components = other_channell; cimage = (j, pim); _; } as z) =
    { z with cimage = (j, Ddgraph.do_it (other_channell, zs) pim); }
  ;;
  
  (* Steps 2 and 3. *)
  (* scalar zonation -> best_component_cutting *)
  let scalar_best_component_cutting z =
    let (_j, pim) = z.cimage in
    pimage_fold_left better_cutting No_more_component_cutting pim
  ;;
  
  (* best_component_cutting -> scalar zonation -> scalar zonation *)
  let scalar_split bcc
      ({ gimage_data = gdata; cimage = (j_i, pim_i); _; } as z) =
    match bcc with
    | No_more_component_cutting -> z
    | Best_component_cutting
        { zone_number = best_zn;
          zone_cutting_candidate =
            { indicator = _delta_j; cutting = (szonet, szonef); }; } ->
        let { Dgraph.value =
              ({ Mapping.domain = best_dom; _; } as gim); _; } =
          Structural_mapping.get pim_i best_zn in
        let { zc_cost = j; _; } = Mapping.get_constant gim in
        let ss = Set.subset_of_full gdata.Mapping.domain best_dom in
        let gdata_bss = Mapping.restriction gdata ss in
        let mean_sub_gimage szone =
          let gdata_szone = Mapping.restriction gdata_bss szone in
          constant_zone Scal.mean gdata_szone Scal.cost_sub gdata in
        let gimt = mean_sub_gimage szonet
        and gimf = mean_sub_gimage szonef in
        let { zc_cost = jt; zc_param = paramt; } = Mapping.get_constant gimt
        and { zc_cost = jf; zc_param = paramf; } = Mapping.get_constant gimf in
        let delta_j = Scal.( -. ) j (Scal.( +. ) jt jf) in
        let pim_ip1 =
          Structural_mapping.set_next
            (Structural_mapping.set pim_i best_zn gimf) gimt in
        print_zone_splitting_report ppf
          best_zn pim_ip1 szonef paramf szonet paramt;
        (*let j_ip1 = Scal.cost pim_ip1 in*)
        let j_ip1 = Scal.( -. ) j_i delta_j in
        { z with cimage = (j_ip1, pim_ip1); }
  ;;
  
  (* best_global_cutting -> component_number -> best_component_cutting ->
     best_global_cutting *)
  let better_component_cutting accu comp_nb bcc =
    match (accu, bcc) with
    | (_, No_more_component_cutting) -> accu
    | (No_more_global_cutting, Best_component_cutting ccc) ->
        bgc_of comp_nb ccc
    | (Best_global_cutting best_gcc, Best_component_cutting ccc) ->
        if indicator_of_ccc ccc <= indicator_of_gcc best_gcc then accu else
        let bgc =
          { component_number = comp_nb;
            component_cutting_candidate = ccc; } in
        Best_global_cutting bgc
  ;;
  
  (* Steps 4 and 5. *)
  (* best_component_cutting codomain ->
     (nat -> nat) * (scalar zonation -> scalar zonation) codomain *)
  let make_dm_cutting bccs =
    let dm = Config.get_dimension_management () in
    let bgc =
      Cod.fold_left_i better_component_cutting
        No_more_global_cutting bccs in
    match bgc with
    | No_more_global_cutting ->
        (Upervasives.id, Cod.init (fun _ -> Upervasives.id))
    | Best_global_cutting gcc ->
        let best_scalar_split =
          scalar_split
            (Best_component_cutting gcc.component_cutting_candidate) in
        let cuttings =
          match dm with
          | (Best_cutting, All_components) -> (* Vector. *)
              Cod.init (fun _ -> best_scalar_split)
          | (Best_cutting, Best_component) -> (* Best_component_only. *)
              Cod.singleton
                gcc.component_number best_scalar_split Upervasives.id
          | (All_cuttings, Best_component) -> (* Best_component_for_each. *)
              Cod.map scalar_split bccs
          | (All_cuttings, All_components) -> (* Combine_best_components. *)
              s_failwith "iteration"
                "Combine_best_components not yet implemented" in
        (Nat.succ, cuttings)
  ;;
  
  (* history -> history *)
  let iteration zsh_i =
    let i = H.head_state_number zsh_i in
    let (St.State zs_i as st_i) = H.head_state zsh_i in
    let zs_i = Cod.map (scalar_cutting_candidates st_i) zs_i in
    let bccs = Cod.map scalar_best_component_cutting zs_i in
    let (succ, cuttings) = make_dm_cutting bccs in
    let ip1 = succ i in
    let zs_ip1 = Cod.map_rails cuttings zs_i in
    let st_ip1 = St.State zs_ip1 in
    let zsh_ip1 = H.make_history_step ip1 st_ip1 zsh_i in
    s_print_debug "iteration" "zsh_ip1" H.print_head_h_state ppf zsh_ip1;
    zsh_ip1
  ;;

  (*let scalar_stop b zsh_i =
    let { H_types.state_number = i; state = _zsi; } = Ulist1.hd zsh_i in
    let best_indic_i = scalar_best_indicator z_i in
    let rel_error_i = scalar_relative_error z_i in
    b &&
    (i >= sc.max_iteration ||
     best_indic_i <= sc.min_best_scalar_indicator ||
     rel_error_i <= sc.min_scalar_relative_error)
  ;;*)
  
  (* history -> bool *)
  let stop zsh_i =
    let sc = Config.get_stopping_criterion () in
    let i = H.head_state_number zsh_i in
    i >= sc.max_iteration
  ;;
  
  (* history -> history *)
  let rec iterations zsh_i =
    Disp.display zsh_i;
    print_iteration_report ppf
      (H.head_state_number zsh_i) (St.cost (H.head_state zsh_i));
    if stop zsh_i then zsh_i else
    let zsh_ip1 = iteration zsh_i in
    let stagnation = H.same_head_state zsh_ip1 zsh_i in
    if stagnation then zsh_i else iterations zsh_ip1
  ;;
  
  (* Algorithm. *)
  
  (* unit -> unit *)
  let algorithm () =
    s_print_debug "algorithm" "zsh_0" H.print_head_h_state ppf zsh_0;
    Disp.init zsh_0;
    let zsh_n = iterations zsh_0 in
    H.finalization zsh_n;
  ;;
end
;;

(*
module Float_single = Make (Scalar.Float) (Codomain.Single);;

module Float_triple = Make (Scalar.Float) (Codomain.Triple);;

module Float_vector =
  functor (V : Vector_spec.VECTOR) ->
  Make (Scalar.Float) (Codomain.Vector (V))
;;
*)
