(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: umatrix.ml,v 1.31 2014/02/09 17:27:28 fclement Exp $ *)

(* Matrices. *)

let m_qual = Upervasives.qualify "Umatrix";;

let m_invalid_arg fun_name = Upervasives.invalid_arg_qual m_qual fun_name "";;

type 'a matrix = 'a Umatrix_types.matrix;;
type 'a t = 'a matrix;;

let of_array_array = Umatrix_types.to_matrix
and to_array_array = Umatrix_types.of_matrix
;;

external unsafe_of_array_array : 'a array array -> 'a matrix = "%identity";;

let _0_ = Unatural._0_;;

let size m =
  let aa = to_array_array m in
  let ni = Array.length aa in
  if ni = 0 then (_0_, _0_) else
  let nj = Array.length (Array.unsafe_get aa 0) in
  (Unatural.of_int ni, Unatural.of_int nj)
;;

let length = size;;

let dim_i m = Uarray.length (to_array_array m);;

let dim_j m = snd (size m);;

let dim1 = dim_i
and dim2 = dim_j
;;

let get m i = Array.get (Array.get (to_array_array m) i)
and set m i = Array.set (Array.get (to_array_array m) i)
;;

let unsafe_get m i = Array.unsafe_get (Array.unsafe_get (to_array_array m) i)
and unsafe_set m i = Array.unsafe_set (Array.unsafe_get (to_array_array m) i)
;;

let get_line m i = Array.get (to_array_array m) i;;

let init ni nj f =
  of_array_array (Uarray.init ni (fun i -> Uarray.init nj (f i)))
;;

let make ni nj x =
  of_array_array (Uarray.init ni (fun _i -> Uarray.make nj x))
;;

let check fun_name f m1 m2 =
  if f m1 <> f m2 then m_invalid_arg fun_name else ()
;;
(** [check s f m1 m2] raises [Invalid_arg] if [f] applied to matrices [m1] and
   [m2] do not return the same result. *)

let append_i m1 m2 =
  check "append_i" dim_j m1 m2;
  of_array_array (Array.append (to_array_array m1) (to_array_array m2))
;;

let append_j m1 m2 =
  check "append_j" dim_i m1 m2;
  of_array_array
    (Uarray.map2 Array.append (to_array_array m1) (to_array_array m2))
;;

let append1 = append_i
and append2 = append_j
;;

let append m1 m2 m3 m4 = append_i (append_j m1 m2) (append_j m3 m4);;

let list_check fun_name f = function
  | [] -> ()
  | x :: l ->
      Ulist.iter (fun y -> if f y <> f x then m_invalid_arg fun_name) l
;;
(** [list_check s f ml] raises [Invalid_arg] if [f] applied to all items of
   the matrix list [ml] do not return the same result. *)

let concat_i ml =
  list_check "concat_i" dim_j ml;
  of_array_array (Array.concat (Ulist.map to_array_array ml))
;;

let concat_j ml =
  list_check "concat_j" dim_i ml;
  of_array_array (Uarray.maplist Array.concat (Ulist.map to_array_array ml))
;;

let concat1 = concat_i
and concat2 = concat_j
;;

let concat mll = concat_i (Ulist.map concat_j mll);;

let sub m oi li oj lj =
  let (ni, nj) = size m in
  if oi < 0 || oi + Unatural.to_int li > Unatural.to_int ni ||
     oj < 0 || oj + Unatural.to_int lj > Unatural.to_int nj then
    m_invalid_arg "sub" else
  init li lj (fun i j -> unsafe_get m (oi + i) (oj + j))
;;

let copy m =
  let (ni, nj) = size m in
  init ni nj (fun i j -> unsafe_get m i j)
;;

let fill m oi li oj lj x =
  let (ni, nj) = size m in
  if oi < 0 || oi + Unatural.to_int li > Unatural.to_int ni ||
     oj < 0 || oj + Unatural.to_int lj > Unatural.to_int nj then
    m_invalid_arg "fill" else
  for i = oi to oi + Unatural.to_int li - 1 do
    for j = oj to oj + Unatural.to_int lj - 1 do
      unsafe_set m i j x
    done;
  done
;;

let blit m1 oi1 oj1 m2 oi2 oj2 li lj =
  let (ni1, nj1) = size m1 and (ni2, nj2) = size m2 in
  if oi1 < 0 || oi1 + Unatural.to_int li > Unatural.to_int ni1 ||
     oj1 < 0 || oj1 + Unatural.to_int lj > Unatural.to_int nj1 ||
     oi2 < 0 || oi2 + Unatural.to_int li > Unatural.to_int ni2 ||
     oj2 < 0 || oj2 + Unatural.to_int lj > Unatural.to_int nj2 then
    m_invalid_arg "blit" else
  let bl i j =
    unsafe_set m2 (oi2 + i) (oj2 + j) (unsafe_get m1 (oi1 + i) (oj1 + j)) in
  if oi1 < oi2 then
    for i = Unatural.to_int li - 1 downto 0 do
      if oj1 < oj2 then
        for j = Unatural.to_int lj - 1 downto 0 do bl i j done
      else
        for j = 0 to Unatural.to_int lj - 1 do bl i j done
    done
  else
    for i = 0 to Unatural.to_int li - 1 do
      if oj1 < oj2 then
        for j = Unatural.to_int lj - 1 downto 0 do bl i j done
      else
        for j = 0 to Unatural.to_int lj - 1 do bl i j done
    done
;;

let transpose m =
  let (ni, nj) = size m in
  if ni == _0_ || nj == _0_ then of_array_array [||] else
  let mt = make nj ni (unsafe_get m 0 0) in
  for j = 0 to Unatural.to_int nj - 1 do
    for i = 0 to Unatural.to_int ni - 1 do
      unsafe_set mt j i (unsafe_get m i j)
    done;
  done;
  mt
;;

let to_array m = Array.concat (Array.to_list (to_array_array m));;

let of_array a ni nj =
  let len = Uarray.length a in
  if len <> Unatural.mul ni nj then m_invalid_arg "of_array" else
  init ni nj (fun i j -> Array.unsafe_get a (Unatural.to_int nj * i + j))
;;

let to_arraylist m = Array.to_list (to_array_array m);;

let of_arraylist al = of_array_array (Array.of_list al);;

let to_list m = Ulist.map Array.to_list (Array.to_list (to_array_array m));;

let of_list l = of_array_array (Array.map Array.of_list (Array.of_list l));;

(** {6 Iterators} *)

let iter f m = Array.iter (Array.iter f) (to_array_array m);;

let map f m = of_array_array (Array.map (Array.map f) (to_array_array m));;

let iter_ij f m =
  Array.iteri (fun i -> Array.iteri (fun j x -> f i j x)) (to_array_array m)
;;

let map_ij f m =
  of_array_array
    (Array.mapi (fun i -> Array.mapi (fun j x -> f i j x)) (to_array_array m))
;;

let iter_i f m = Array.iter f (to_array_array m);;

let iter_j f m = iter_i f (transpose m);;

let map_i f m = of_array_array (Array.map f (to_array_array m));;

let map_j f m = transpose (map_i f (transpose m));;

let fold_left f accu m =
  Array.fold_left (Array.fold_left f) accu (to_array_array m)
;;

let fold_right f m =
  Array.fold_right (Array.fold_right f) (to_array_array m)
;;

let sub_iter f m oi li oj lj =
  Uarray.sub_iter (fun mi -> Uarray.sub_iter f mi oj lj) (to_array_array m) oi li
;;

let sub_iter_ij f m oi li oj lj =
  Uarray.sub_iter_i
    (fun i mi -> Uarray.sub_iter_i (f i) mi oj lj)
    (to_array_array m) oi li
;;

(** {6 Iterators on two matrices} *)

let iter2 f m1 m2 =
  Uarray.iter2 (Uarray.iter2 f) (to_array_array m1) (to_array_array m2)
;;

let map2 f m1 m2 =
  of_array_array
    (Uarray.map2 (Uarray.map2 f) (to_array_array m1) (to_array_array m2))
;;

let iter_ij2 f m1 m2 =
  Uarray.iter_i2
    (fun i -> Uarray.iter_i2 (f i))
    (to_array_array m1) (to_array_array m2)
;;

let map_ij2 f m1 m2 =
  of_array_array
    (Uarray.map_i2
       (fun i -> Uarray.map_i2 (f i))
       (to_array_array m1) (to_array_array m2))
;;

let fold_left2 f accu m1 m2 =
  Uarray.fold_left2
    (Uarray.fold_left2 f)
    accu (to_array_array m1) (to_array_array m2)
;;

let fold_right2 f m1 m2 =
  Uarray.fold_right2
    (Uarray.fold_right2 f)
    (to_array_array m1) (to_array_array m2)
;;

(* Matrix scanning. *)

let for_all p m = Uarray.for_all (Uarray.for_all p) (to_array_array m);;

(* Specific applications. *)

let make_diagonal_matrix zero v =
  let n = Uarray.length v in
  let f i j = if i = j then v.(i) else zero in
  init n n f
;;

let make_int_diagonal_matrix v = make_diagonal_matrix 0 v;;
let make_float_diagonal_matrix v = make_diagonal_matrix 0. v;;

let resize ni2 nj2 m1 =
  let (ni1, nj1) = size m1 in
  let aa1 = to_array_array m1 in
  match (Unatural.to_int ni1, Unatural.to_int nj1,
         Unatural.to_int ni2, Unatural.to_int nj2) with
  | (0, _, 0, _) -> of_array_array [||]
  | (_, 0, _, 0) -> of_array_array (Uarray.make ni2 (Array.unsafe_get aa1 0))
  | (0, _, _, _) | (_, 0, _, _) -> m_invalid_arg "resize"
  | (_, _, _, _) ->
    let m2 = make ni2 nj2 (unsafe_get m1 0 0) in
    let fi i = (i * Unatural.to_int ni2) / Unatural.to_int ni1
    and fj j = (j * Unatural.to_int nj2) / Unatural.to_int nj1 in
    iter_ij
      (fun i1 j1 x ->
        for i2 = fi i1 to pred (fi (succ i1)) do
        for j2 = fj j1 to pred (fj (succ j1)) do
          unsafe_set m2 i2 j2 x;
        done;
        done)
      m1;
    m2
;;

let zoom ni nj m =
  let (ni1, nj1) = size m in
  resize (Unatural.mul ni ni1) (Unatural.mul nj nj1) m
;;
