(***********************************************************************)
(*                                                                     *)
(*                                Ulib                                 *)
(*                                                                     *)
(*                   OCaml user's additional library                   *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2004-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ustamped_darray.mli,v 1.8 2014/01/16 13:35:38 fclement Exp $ *)

(** {3 Stamped dynamic arrays} *)

(** Stamped dynamic arrays can be filled incrementally without keeping unset
 holes. Elements are not valid until set once. Already set elements can be set
 again, but cannot be unset. *)

type index = Udynamic_array.index;;
type length = Udynamic_array.length;;

type 'a stamped =
  | Unset
  | Valid of 'a
;;
(** Elements of stamped dynamic arrays are either unset or valid. *)

exception Not_set;;

type 'a t =
  {
   darray : 'a stamped Udynamic_array.t;
   mutable next_unset : index;
  }
;;

val make : length -> 'a t;;

val get : 'a t -> index -> 'a;;
(** [get sda i] returns item number [i] of stamped dynamic array [sda].

 Raise [Not_set] if item number [i] is not set. *)

val set_next : 'a t -> 'a -> unit;;

val set : 'a t -> index -> 'a -> unit;;
(** Raise [Invalid_arg] if the argument index is greater than next unset. *)

val same : 'a t -> 'a t -> bool;;

val max_length : length;;
val length : 'a t -> length;;

(** {6 Converters} *)

val to_array : 'a t -> 'a array;;
(** [to_array sda] returns the array of valid values. *)

val of_array : 'a array -> 'a t;;

(** {6 Iterators} *)

val iter : (index -> 'a -> unit) -> 'a t -> unit;;
val map : (index -> 'a -> 'b) -> 'a t -> 'b t;;
val fold_left : ('accu -> index -> 'a -> 'accu) -> 'accu -> 'a t -> 'accu;;
